@@ -22,9 +22,7 @@ from typing import (
 )
 from typing_extensions import Literal
 
-import aiohttp
 import attr
-from async_timeout import timeout
 import zmq
 
 from aiodocker.docker import Docker, DockerContainer
@@ -72,14 +70,12 @@ from ..resources import (
 from ..server import (
     get_extra_volumes,
 )
-from ..stats import (
-    spawn_stat_synchronizer, StatSyncState
-)
 from ..types import (
     Container, Port, ContainerStatus,
     ContainerLifecycleEvent, LifecycleEvent,
 )
 from ..utils import (
+    current_loop,
     update_nested_dict,
     get_kernel_id_from_container,
     host_pid_to_container_pid,
@@ -114,8 +110,7 @@ def container_from_docker_container(src: DockerContainer) -> Container:
 class DockerAgent(AbstractAgent):
 
     docker: Docker
-    monitor_fetch_task: asyncio.Task
-    monitor_handle_task: asyncio.Task
+    monitor_docker_task: asyncio.Task
     agent_sockpath: Path
     agent_sock_task: asyncio.Task
     scan_images_timer: asyncio.Task
@@ -131,23 +126,16 @@ class DockerAgent(AbstractAgent):
         await super().__ainit__()
         self.agent_sockpath = ipc_base_path / f'agent.{self.agent_id}.sock'
         self.agent_sock_task = self.loop.create_task(self.handle_agent_socket())
-        self.monitor_fetch_task  = self.loop.create_task(self.fetch_docker_events())
-        self.monitor_handle_task = self.loop.create_task(self.handle_docker_events())
+        self.monitor_docker_task = self.loop.create_task(self.monitor_docker_events())
 
     async def shutdown(self, stop_signal: signal.Signals):
         try:
             await super().shutdown(stop_signal)
         finally:
             # Stop docker event monitoring.
-            if self.monitor_fetch_task is not None:
-                self.monitor_fetch_task.cancel()
-                self.monitor_handle_task.cancel()
-                await self.monitor_fetch_task
-                await self.monitor_handle_task
-            try:
-                await self.docker.events.stop()
-            except Exception:
-                pass
+            if self.monitor_docker_task is not None:
+                self.monitor_docker_task.cancel()
+                await self.monitor_docker_task
             await self.docker.close()
 
         # Stop handlign agent sock.
@@ -478,6 +466,11 @@ class DockerAgent(AbstractAgent):
             nonlocal mounts
             mounts.append(Mount(type, src, target, MountPermission(perm), opts=opts))
 
+        async def _copy(src: Union[str, Path], target: Union[str, Path]):
+            def sync():
+                shutil.copy(src, target)
+            await self.loop.run_in_executor(None, sync)
+
         # Inject Backend.AI kernel runner dependencies.
         distro = image_labels.get('ai.backend.base-distro', 'ubuntu16.04')
         matched_distro, krunner_volume = match_krunner_volume(
@@ -580,21 +573,16 @@ class DockerAgent(AbstractAgent):
         # we need to touch them first to avoid their "ghost" files are created
         # as root in the host-side filesystem, which prevents deletion of scratch
         # directories when the agent is running as non-root.
-        (work_dir / '.jupyter' / 'custom').mkdir(parents=True, exist_ok=True)
-        (work_dir / '.jupyter' / 'custom' / 'custom.css').write_bytes(b'')
-        (work_dir / '.jupyter' / 'custom' / 'logo.svg').write_bytes(b'')
-        (work_dir / '.jupyter' / 'custom' / 'roboto.ttf').write_bytes(b'')
-        (work_dir / '.jupyter' / 'custom' / 'roboto-italic.ttf').write_bytes(b'')
-        _mount(MountTypes.BIND, jupyter_custom_css_path.resolve(),
-                                '/home/work/.jupyter/custom/custom.css')
-        _mount(MountTypes.BIND, logo_path.resolve(), '/home/work/.jupyter/custom/logo.svg')
-        _mount(MountTypes.BIND, font_path.resolve(), '/home/work/.jupyter/custom/roboto.ttf')
-        _mount(MountTypes.BIND, font_italic_path.resolve(),
-                                '/home/work/.jupyter/custom/roboto-italic.ttf')
-        _mount(MountTypes.BIND, bashrc_path.resolve(), '/home/work/.bashrc')
-        _mount(MountTypes.BIND, bash_profile_path.resolve(), '/home/work/.bash_profile')
-        _mount(MountTypes.BIND, vimrc_path.resolve(), '/home/work/.vimrc')
-        _mount(MountTypes.BIND, tmux_conf_path.resolve(), '/home/work/.tmux.conf')
+        jupyter_custom_dir = (work_dir / '.jupyter' / 'custom')
+        jupyter_custom_dir.mkdir(parents=True, exist_ok=True)
+        await _copy(jupyter_custom_css_path.resolve(), jupyter_custom_dir / 'custom.css')
+        await _copy(logo_path.resolve(), jupyter_custom_dir / 'logo.svg')
+        await _copy(font_path.resolve(), jupyter_custom_dir / 'roboto.ttf')
+        await _copy(font_italic_path.resolve(), jupyter_custom_dir / 'roboto-italic.ttf')
+        await _copy(bashrc_path.resolve(), work_dir / '.bashrc')
+        await _copy(bash_profile_path.resolve(), work_dir / '.bash_profile')
+        await _copy(vimrc_path.resolve(), work_dir / '.vimrc')
+        await _copy(tmux_conf_path.resolve(), work_dir / '.tmux.conf')
         environ['LD_PRELOAD'] = '/opt/kernel/libbaihook.so'
         if self.config['debug']['coredump']['enabled']:
             _mount(MountTypes.BIND, self.config['debug']['coredump']['path'],
@@ -659,9 +647,10 @@ class DockerAgent(AbstractAgent):
                     os.chown(work_dir, uid, gid)
                     os.chown(work_dir / '.jupyter', uid, gid)
                     os.chown(work_dir / '.jupyter' / 'custom', uid, gid)
-                    os.chown(bashrc_path, uid, gid)
-                    os.chown(bash_profile_path, uid, gid)
-                    os.chown(vimrc_path, uid, gid)
+                    os.chown(work_dir / '.bashrc', uid, gid)
+                    os.chown(work_dir / '.bash_profile', uid, gid)
+                    os.chown(work_dir / '.vimrc', uid, gid)
+                    os.chown(work_dir / '.tmux.conf', uid, gid)
             os.makedirs(config_dir, exist_ok=True)
             # Store custom environment variables for kernel runner.
             with open(config_dir / 'kconfig.dat', 'wb') as fb:
@@ -685,6 +674,16 @@ class DockerAgent(AbstractAgent):
             docker_creds = internal_data.get('docker_credentials')
             if docker_creds:
                 (config_dir / 'docker-creds.json').write_text(json.dumps(docker_creds))
+            # Create bootstrap.sh into workdir if needed
+            bootstrap = internal_data.get('bootstrap_script')
+            sys.stdout.flush()
+            if bootstrap:
+                (work_dir / 'bootstrap.sh').write_text(bootstrap)
+                if KernelFeatures.UID_MATCH in kernel_features:
+                    uid = self.config['container']['kernel-uid']
+                    gid = self.config['container']['kernel-gid']
+                    if os.geteuid() == 0:
+                        os.chown(work_dir / 'bootstrap.sh', uid, gid)
 
         # Create SSH keypair only if ssh_keypair internal_data exists and
         # /home/work/.ssh folder is not mounted.
@@ -852,13 +851,7 @@ class DockerAgent(AbstractAgent):
                     for k, v in kvpairs.items():
                         f.write(f'{k}={v}\n')
 
-            stat_sync_state = StatSyncState(kernel_id)
-            self.stat_sync_states[cid] = stat_sync_state
-            async with spawn_stat_synchronizer(self.config['_src'],
-                                               self.stat_sync_sockpath,
-                                               self.stat_ctx.mode, cid) as proc:
-                stat_sync_state.sync_proc = proc
-                await container.start()
+            await container.start()
 
             # Get attached devices information (including model_name).
             attached_devices = {}
@@ -933,7 +926,7 @@ class DockerAgent(AbstractAgent):
                       (kernel_config['startup_command'] or '')[:60])
 
             # TODO: make this working after agent restarts
-            async def execute_batch():
+            async def execute_batch() -> None:
                 opts = {
                     'exec': kernel_config['startup_command'],
                 }
@@ -966,13 +959,16 @@ class DockerAgent(AbstractAgent):
                         await self.produce_event(
                             'kernel_failure', str(kernel_id), -2)
                         break
+                    opts = {
+                        'exec': '',
+                    }
                 # TODO: store last_stat?
                 destroyed = asyncio.Event()
-                self.container_lifecycle_events.put(
+                await self.container_lifecycle_queue.put(
                     ContainerLifecycleEvent(
                         kernel_id,
                         kernel_obj['container_id'],
-                        LifecycleEvent.TERMINATED,
+                        LifecycleEvent.DESTROY,
                         'task-finished',
                         destroyed,
                     )
@@ -1001,28 +997,15 @@ class DockerAgent(AbstractAgent):
     ) -> Optional[Mapping[MetricKey, MetricValue]]:
         if container_id is None:
             return None
+        if kernel_id not in self.restarting_kernels:
+            self.terminating_kernels.add(kernel_id)
         try:
             container = self.docker.containers.container(container_id)
-            await container.kill()
-            # Collect the last-moment statistics.
-            if container_id in self.stat_sync_states:
-                s = self.stat_sync_states[container_id]
-                try:
-                    with timeout(5):
-                        await s.terminated.wait()
-                        if s.sync_proc is not None:
-                            await s.sync_proc.wait()
-                            s.sync_proc = None
-                except ProcessLookupError:
-                    pass
-                except asyncio.TimeoutError:
-                    log.warning('stat-collector shutdown sync timeout.')
-                last_stat: MutableMapping[MetricKey, MetricValue] = {
-                    key: metric.to_serializable_dict()
-                    for key, metric in s.last_stat.items()
-                }
-                last_stat['version'] = 2  # type: ignore
-                return last_stat
+            # The default timeout of the docker stop API is 10 seconds
+            # to kill if container does not self-terminate.
+            await container.stop()
+            # TODO: last_stat must be collected in the manager from Redis/DB.
+            return None
         except DockerError as e:
             if e.status == 409 and 'is not running' in e.message:
                 # already dead
@@ -1044,24 +1027,12 @@ class DockerAgent(AbstractAgent):
         container_id: Optional[ContainerId],
         restarting: bool,
     ) -> None:
-        if container_id is not None:
-            stat_sync_state = self.stat_sync_states.pop(container_id, None)
-            if stat_sync_state:
-                sync_proc = stat_sync_state.sync_proc
-                try:
-                    if sync_proc is not None:
-                        sync_proc.terminate()
-                        try:
-                            with timeout(2.0):
-                                await sync_proc.wait()
-                        except asyncio.TimeoutError:
-                            sync_proc.kill()
-                            await sync_proc.wait()
-                except ProcessLookupError:
-                    pass
+        loop = current_loop()
 
         kernel_obj = self.kernel_registry.get(kernel_id)
         if kernel_obj is not None:
+            if kernel_id not in self.restarting_kernels:
+                self.terminating_kernels.add(kernel_id)
             for domain_socket_proxy in kernel_obj.get('domain_socket_proxies', []):
                 domain_socket_proxy.proxy_server.close()
                 await domain_socket_proxy.proxy_server.wait_closed()
@@ -1072,103 +1043,81 @@ class DockerAgent(AbstractAgent):
 
         # When the agent restarts with a different port range, existing
         # containers' host ports may not belong to the new port range.
-        if not self.config['debug']['skip-container-deletion'] and container_id is not None:
-            container = self.docker.containers.container(container_id)
-            try:
-                with timeout(20):
-                    await container.delete()
-            except DockerError as e:
-                if e.status == 409 and 'already in progress' in e.message:
-                    pass
-                elif e.status == 404:
-                    pass
-                else:
-                    log.exception(
-                        'unexpected docker error while deleting container (k:{}, c:{})',
-                        kernel_id, container_id)
-            except asyncio.TimeoutError:
-                log.warning('container deletion timeout (k:{}, c:{})',
+        try:
+            if not self.config['debug']['skip-container-deletion'] and container_id is not None:
+                container = self.docker.containers.container(container_id)
+                try:
+                    await container.delete(force=True, v=True)
+                except DockerError as e:
+                    if e.status == 409 and 'already in progress' in e.message:
+                        pass
+                    elif e.status == 404:
+                        pass
+                    else:
+                        log.exception(
+                            'unexpected docker error while deleting container (k:{}, c:{})',
                             kernel_id, container_id)
+                except asyncio.TimeoutError:
+                    log.warning('container deletion timeout (k:{}, c:{})',
+                                kernel_id, container_id)
 
-        if not restarting:
-            scratch_root = self.config['container']['scratch-root']
-            scratch_dir = scratch_root / str(kernel_id)
-            tmp_dir = scratch_root / f'{kernel_id}_tmp'
-            try:
-                if (sys.platform.startswith('linux') and
-                    self.config['container']['scratch-type'] == 'memory'):
-                    await destroy_scratch_filesystem(scratch_dir)
-                    await destroy_scratch_filesystem(tmp_dir)
-                    shutil.rmtree(tmp_dir)
-                shutil.rmtree(scratch_dir)
-            except FileNotFoundError:
-                pass
+            if not restarting:
+                scratch_root = self.config['container']['scratch-root']
+                scratch_dir = scratch_root / str(kernel_id)
+                tmp_dir = scratch_root / f'{kernel_id}_tmp'
+                try:
+                    if (sys.platform.startswith('linux') and
+                        self.config['container']['scratch-type'] == 'memory'):
+                        await destroy_scratch_filesystem(scratch_dir)
+                        await destroy_scratch_filesystem(tmp_dir)
+                        await loop.run_in_executor(None, shutil.rmtree, tmp_dir)
+                    await loop.run_in_executor(None, shutil.rmtree, scratch_dir)
+                except FileNotFoundError:
+                    pass
+        finally:
+            self.terminating_kernels.discard(kernel_id)
 
-    async def fetch_docker_events(self):
-        while True:
-            try:
-                await self.docker.events.run()
-            except asyncio.TimeoutError:
-                # The API HTTP connection may terminate after some timeout
-                # (e.g., 5 minutes)
-                log.info('restarting docker.events.run()')
-                continue
-            except aiohttp.ClientError as e:
-                log.warning('restarting docker.events.run() due to {0!r}', e)
-                continue
-            except asyncio.CancelledError:
-                break
-            except Exception:
-                log.exception('unexpected error')
-                self.error_monitor.capture_exception()
-                break
-
-    async def handle_docker_events(self):
-        subscriber = self.docker.events.subscribe()
-        last_footprint = None
+    async def monitor_docker_events(self):
+        subscriber = self.docker.events.subscribe(create_task=True)
         while True:
             try:
                 evdata = await subscriber.get()
+                if self.config['debug']['log-docker-events']:
+                    log.debug('docker-event: raw: {}', evdata)
+                if evdata['Action'] == 'start':
+                    container_name = evdata['Actor']['Attributes']['name']
+                    kernel_id = await get_kernel_id_from_container(container_name)
+                    if kernel_id is None:
+                        continue
+                    await self.inject_container_lifecycle_event(
+                        kernel_id,
+                        LifecycleEvent.START,
+                        'new-container-started',
+                        container_id=ContainerId(evdata['Actor']['ID']),
+                    )
+                elif evdata['Action'] == 'die':
+                    # When containers die, we immediately clean up them.
+                    container_name = evdata['Actor']['Attributes']['name']
+                    kernel_id = await get_kernel_id_from_container(container_name)
+                    if kernel_id is None:
+                        continue
+                    reason = None
+                    kernel_obj = self.kernel_registry.get(kernel_id)
+                    if kernel_obj is not None:
+                        reason = kernel_obj.termination_reason
+                    try:
+                        exit_code = evdata['Actor']['Attributes']['exitCode']
+                    except KeyError:
+                        exit_code = 255
+                    await self.inject_container_lifecycle_event(
+                        kernel_id,
+                        LifecycleEvent.CLEAN,
+                        reason or 'self-terminated',
+                        container_id=ContainerId(evdata['Actor']['ID']),
+                        exit_code=exit_code,
+                    )
             except asyncio.CancelledError:
                 break
-            if evdata is None:
-                # fetch_docker_events() will automatically reconnect.
-                continue
-
-            # FIXME: Sometimes(?) duplicate event data is received.
-            # Just ignore the duplicate ones.
-            new_footprint = (
-                evdata['Type'],
-                evdata['Action'],
-                evdata['Actor']['ID'],
-            )
-            if new_footprint == last_footprint:
-                continue
-            last_footprint = new_footprint
-
-            if self.config['debug']['log-docker-events']:
-                log.debug('docker-event: raw: {}', evdata)
-
-            if evdata['Action'] == 'die':
-                # When containers die, we immediately clean up them.
-                container_name = evdata['Actor']['Attributes']['name']
-                kernel_id = await get_kernel_id_from_container(container_name)
-                if kernel_id is None:
-                    continue
-                reason = None
-                kernel_obj = self.kernel_registry.get(kernel_id)
-                if kernel_obj is not None:
-                    reason = kernel_obj.termination_reason
-                try:
-                    exit_code = evdata['Actor']['Attributes']['exitCode']
-                except KeyError:
-                    exit_code = 255
-                await self.inject_container_lifecycle_event(
-                    kernel_id,
-                    LifecycleEvent.CLEAN,
-                    reason or 'self-terminated',
-                    container_id=ContainerId(evdata['Actor']['ID']),
-                    exit_code=exit_code,
-                )
-
-        await asyncio.sleep(0.5)
+            except Exception:
+                log.exception('unexpected error while processing docker events')
+        await self.docker.events.stop()
