@@ -43,6 +43,8 @@ if TYPE_CHECKING:
 
 log = BraceStyleAdapter(logging.getLogger('ai.backend.agent.server'))
 
+agent_instance = None  # for live-debugging
+
 redis_config_iv = t.Dict({
     t.Key('addr', default=('127.0.0.1', 6379)): tx.HostPortPair,
     t.Key('password', default=None): t.Null | t.String,
@@ -196,10 +198,10 @@ class AgentRPCServer(aiozmq.rpc.AttrHandler, aobject):
         except AssertionError:
             log.exception('assertion failure')
             raise
-        except Exception:
-            log.exception('unexpected error')
+        except Exception as e:
+            log.exception('unhandled error')
             self.error_monitor.capture_exception()
-            raise
+            raise RuntimeError('unhandled error during rpc call', repr(e))
 
     @aiozmq.rpc.method
     def ping(self, msg: str) -> str:
@@ -288,10 +290,10 @@ class AgentRPCServer(aiozmq.rpc.AttrHandler, aobject):
                       flush_timeout,      # type: float
                       ):
         # type: (...) -> Dict[str, Any]
-        _log = log.debug if mode == 'continue' else log.info
-        _log('rpc::execute(k:{0}, run-id:{1}, mode:{2}, code:{3!r})',
-             kernel_id, run_id, mode,
-             code[:20] + '...' if len(code) > 20 else code)
+        if mode != 'continue':
+            log.info('rpc::execute(k:{0}, run-id:{1}, mode:{2}, code:{3!r})',
+                     kernel_id, run_id, mode,
+                     code[:20] + '...' if len(code) > 20 else code)
         async with self.handle_rpc_exception():
             result = await self.agent.execute(
                 KernelId(UUID(kernel_id)),
@@ -450,7 +452,9 @@ async def server_main(loop, pidx, _args):
 
     # Start RPC server.
     try:
+        global agent_instance
         agent = await AgentRPCServer.new(etcd, config)
+        agent_instance = agent
     except InitializationError as e:
         log.error('Agent initialization failed: {}', e)
         os.kill(0, signal.SIGINT)
