@@ -13,7 +13,7 @@ from typing import (
     ClassVar,
     Optional, Union,
     Sequence,
-    MutableMapping,
+    MutableMapping, Dict,
 )
 
 import janus
@@ -29,6 +29,7 @@ from .utils import wait_local_port_open
 from .intrinsic import (
     init_sshd_service, prepare_sshd_service,
     prepare_ttyd_service,
+    prepare_vscode_service,
 )
 
 log = BraceStyleAdapter(logging.getLogger())
@@ -84,6 +85,8 @@ class BaseRunner(metaclass=ABCMeta):
     subproc: Optional[asyncio.subprocess.Process]
     runtime_path: Optional[Path]
 
+    services_running: Dict[str, asyncio.subprocess.Process]
+
     def __init__(self, loop=None):
         self.subproc = None
         self.runtime_path = None
@@ -102,7 +105,6 @@ class BaseRunner(metaclass=ABCMeta):
             log.exception('Reading /home/config/environ.txt failed!')
 
         # initialized after loop creation
-        self.loop = loop if loop is not None else current_loop()
         self.zctx = zmq.asyncio.Context()
         self.started_at: float = time.monotonic()
         self.insock = None
@@ -111,8 +113,7 @@ class BaseRunner(metaclass=ABCMeta):
         self.task_queue = None
         self.log_queue = None
 
-        self.services_running = set()
-        self.service_processes = []
+        self.services_running = {}
 
         # If the subclass implements interatcive user inputs, it should set a
         # asyncio.Queue-like object to self.user_input_queue in the
@@ -239,7 +240,7 @@ class BaseRunner(metaclass=ABCMeta):
             elif exec_cmd == '*':
                 ret = await self.execute_heuristic()
             else:
-                ret = await self.run_subproc(exec_cmd)
+                ret = await self.run_subproc(exec_cmd, batch=True)
         except Exception:
             log.exception('unexpected error')
             ret = -1
@@ -417,11 +418,18 @@ class BaseRunner(metaclass=ABCMeta):
             if service_info['name'] in self.services_running:
                 result = {'status': 'running'}
                 return
-            print(f"starting service {service_info['name']}")
+            if service_info['protocol'] == 'pty':
+                result = {'status': 'failed',
+                          'error': 'not implemented yet'}
+                return
+            cwd = Path.cwd()
+            cmdargs = None, {}
             if service_info['name'] == 'ttyd':
                 cmdargs, env = await prepare_ttyd_service(service_info)
             elif service_info['name'] == 'sshd':
                 cmdargs, env = await prepare_sshd_service(service_info)
+            elif service_info['name'] == 'vscode':
+                cmdargs, env = await prepare_vscode_service(service_info)
             else:
                 cmdargs, env = await self.start_service(service_info)
             if cmdargs is None:
@@ -430,23 +438,32 @@ class BaseRunner(metaclass=ABCMeta):
                 result = {'status': 'failed',
                           'error': 'unsupported service'}
                 return
-            if service_info['protocol'] == 'pty':
-                # TODO: handle pseudo-tty
-                raise NotImplementedError
-            else:
-                service_env = {**self.child_env, **env}
-                # avoid conflicts with Python binary used by service apps.
-                if 'LD_LIBRARY_PATH' in service_env:
-                    service_env['LD_LIBRARY_PATH'] = \
-                        service_env['LD_LIBRARY_PATH'].replace('/opt/backend.ai/lib:', '')
+            log.debug('cmdargs: {0}', cmdargs)
+            log.debug('env: {0}', env)
+            service_env = {**self.child_env, **env}
+            # avoid conflicts with Python binary used by service apps.
+            if 'LD_LIBRARY_PATH' in service_env:
+                service_env['LD_LIBRARY_PATH'] = \
+                    service_env['LD_LIBRARY_PATH'].replace('/opt/backend.ai/lib:', '')
+            loop = current_loop()
+            try:
                 proc = await asyncio.create_subprocess_exec(
                     *cmdargs,
                     env=service_env,
+                    cwd=cwd,
                 )
-                self.service_processes.append(proc)
-            self.services_running.add(service_info['name'])
-            await wait_local_port_open(service_info['port'])
-            result = {'status': 'started'}
+                self.services_running[service_info['name']] = proc
+                loop.create_task(self._wait_service_proc(service_info['name'], proc))
+                await wait_local_port_open(service_info['port'])
+                log.info("Service {} has started (pid: {}, port: {})",
+                         service_info['name'], proc.pid, service_info['port'])
+                result = {'status': 'started'}
+            except PermissionError:
+                result = {'status': 'failed',
+                          'error': f"the target file is not executable: {cmdargs[0]}"}
+            except FileNotFoundError:
+                result = {'status': 'failed',
+                          'error': f"the executable file is not found: {cmdargs[0]}"}
         except Exception as e:
             log.exception('unexpected error')
             result = {'status': 'failed', 'error': repr(e)}
@@ -456,7 +473,16 @@ class BaseRunner(metaclass=ABCMeta):
                 json.dumps(result).encode('utf8'),
             ])
 
-    async def run_subproc(self, cmd: Sequence[Union[str, Path, None]]):
+    async def _wait_service_proc(
+        self,
+        service_name: str,
+        proc: asyncio.subprocess.Process,
+    ) -> None:
+        exitcode = await proc.wait()
+        log.info(f"Service {service_name} (pid: {proc.pid}) has terminated with exit code: {exitcode}")
+        self.services_running.pop(service_name, None)
+
+    async def run_subproc(self, cmd: Sequence[Union[str, Path, None]], batch: bool = False):
         """A thin wrapper for an external command."""
         loop = current_loop()
         if Path('/home/work/.logs').is_dir():
@@ -479,8 +505,11 @@ class BaseRunner(metaclass=ABCMeta):
             pipe_opts['stdout'] = asyncio.subprocess.PIPE
             pipe_opts['stderr'] = asyncio.subprocess.PIPE
             with open(log_path, 'ab') as log_out:
+                env = {**self.child_env}
+                if batch:
+                    env['_BACKEND_BATCH_MODE'] = '1'
                 proc = await exec_func(
-                    env=self.child_env,
+                    env=env,
                     stdin=None,
                     **pipe_opts,
                 )
@@ -586,7 +615,8 @@ class BaseRunner(metaclass=ABCMeta):
                     await self._send_status()
                 elif op_type == 'start-service':  # activate a service port
                     data = json.loads(text)
-                    await self._start_service(data)
+                    async with self._service_lock:
+                        await self._start_service(data)
             except asyncio.CancelledError:
                 break
             except NotImplementedError:
@@ -602,24 +632,26 @@ class BaseRunner(metaclass=ABCMeta):
 
     async def _init(self, cmdargs):
         self.cmdargs = cmdargs
-        self.loop = current_loop()
         # Initialize event loop.
+        loop = current_loop()
         executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
-        self.loop.set_default_executor(executor)
+        loop.set_default_executor(executor)
+        self.loop = loop
 
-        self.insock = self.zctx.socket(zmq.PULL, io_loop=self.loop)
+        self.insock = self.zctx.socket(zmq.PULL)
         self.insock.bind('tcp://*:2000')
-        self.outsock = self.zctx.socket(zmq.PUSH, io_loop=self.loop)
+        self.outsock = self.zctx.socket(zmq.PUSH)
         self.outsock.bind('tcp://*:2001')
 
-        self.log_queue = janus.Queue(loop=self.loop)
-        self.task_queue = asyncio.Queue(loop=self.loop)
-        self.init_done = asyncio.Event(loop=self.loop)
+        self.log_queue = janus.Queue()
+        self.task_queue = asyncio.Queue()
+        self.init_done = asyncio.Event()
 
         setup_logger(self.log_queue.sync_q, self.log_prefix, cmdargs.debug)
-        self._log_task = self.loop.create_task(self._handle_logs())
-        self._main_task = self.loop.create_task(self.main_loop(cmdargs))
-        self._run_task = self.loop.create_task(self.run_tasks())
+        self._log_task = loop.create_task(self._handle_logs())
+        self._main_task = loop.create_task(self.main_loop(cmdargs))
+        self._run_task = loop.create_task(self.run_tasks())
+        self._service_lock = asyncio.Lock()
 
     async def _shutdown(self):
         try:
@@ -629,12 +661,14 @@ class BaseRunner(metaclass=ABCMeta):
             self._main_task.cancel()
             await self._run_task
             await self._main_task
-            if self.service_processes:
-                log.debug('terminating service processes...')
+            log.debug('terminating service processes...')
+            running_procs = [*self.services_running.values()]
+            async with self._service_lock:
                 await asyncio.gather(
-                    *(terminate_and_kill(proc) for proc in self.service_processes),
+                    *(terminate_and_kill(proc) for proc in running_procs),
                     return_exceptions=True,
                 )
+                await asyncio.sleep(0.01)
             log.debug('terminated.')
         finally:
             # allow remaining logs to be flushed.
